"""
Expression Classes
"""
# Third party imports
from pyomo.core.expr.numeric_expr import (DivisionExpression,
                                          NegationExpression,
                                          ProductExpression)
from pyomo.environ import units as pyo_units

# KIPET library imports
from kipet.model_tools.replacement import _update_expression


class ExpressionBlock:
    
    """Class for general expression block classes

    :Methods:

        - :func:`display`
        - :func:`display_units`
    """
    
    def __init__(self, exprs=None):
        
        self.exprs = exprs
        self._title = 'EXPR'

    def __len__(self):
        
        return len(self.exprs)
        
    def display(self):
        """This method loops through all expressions and displays them in a readable format

        :return: None

        """
        if self.exprs is not None:
            print(f'{self._title} expressions:')
            for key, expr in self.exprs.items():
                print(f'{key}: {expr.expression.to_string()}')

        return None

    def display_units(self):
        """Loops through the expressions and shows their units

        :return: None

        """
        margin = 8
        if self.exprs is not None:
            print(f'{self._title} units:')
            for key, expr in self.exprs.items():
                print(f'{str(key).rjust(margin)} : {expr.units}')

        return None


class ODEExpressions(ExpressionBlock):
    
    """Class for ODE expressions

    :param dict ode_exprs: dict of ODE expressions

    """
    
    def __init__(self, ode_exprs=None):
        
        super().__init__(ode_exprs)
        self._title = 'ODE'
        
        
class AEExpressions(ExpressionBlock):
    
    """Class for AE expressions

    :param dict alg_exprs: dict of alebraic expressions

    """
    
    def __init__(self, alg_exprs=None):
        
        super().__init__(alg_exprs)
        self._title = 'ALG'


class Expression:
    
    """Class for individual expressions

    :param str name: The name of the expression
    :param expression: Pyomo expression

    :Methods:

        - :func:`show_units`
        - :func:`check_units`
        - :func:`check_division`
        - :func:`check_expression_units`

    """
    
    def __init__(self,
                 name,
                 expression):
        
        self.name = name
        self.expression = expression
        self.units = None
        self.expression_orig = None
        
    def __str__(self):
        
        return self.expression.to_string()
    
    @property
    def show_units(self):
        """Display the units of the expression

        :return: String of the units
        :rtype: str

        """
        return self.units.to_string()
    
    def _change_to_unit(self, c_mod, c_mod_new):
        """Method to remove the fixed parameters from the ConcreteModel

        :param Comp c_mod: The Comp object of dummy pyomo variables
        :param Comp c_mod_new: The Comp object of new variables

        """
        var_dict = c_mod_new.var_dict
        expr_new = self.expression
        if self.expression_orig is not None:
            expr_new = self.expression_orig    
        for model_var, var_obj in var_dict.items():
            old_var = c_mod[model_var][1]
            new_var = var_dict[model_var]
            expr_new = _update_expression(expr_new, old_var, new_var)
        return expr_new

    def check_units(self):
        """Check the expr units by exchanging the real model with unit model
        components
        
        :return: None

        """
        self.units = pyo_units.get_units(self.expression)
        return None

    def check_division(self, eps=1e-12):
        """Add a small amount to the numerator and denominator in a
        DivisionExpression to improve the numerics.

        :param float eps: The small addition to the numerator and denominator

        :return: None

        """
        expr = self.expression
        
        if isinstance(expr, DivisionExpression):
        
            ex1, ex2 = expr.args
            ex1 += eps
            ex2 += eps
            expr_new = ex1/ex2
            
            self.expression = expr_new
            self.expression_orig = expr
    
        return None

    @staticmethod
    def _check_term(term, convert_to):
        """This loops through terms in the expression to ensure the units are valid

        :param expression term: The term to check
        :param str convert_to: The units to convert to, if necessary

        :return term_new: The updated expression term
        :rtype: expression

        """
        unit_term = pyo_units.get_units(term)
        term_new = term
        
        if unit_term is not None:
            term_new = pyo_units.convert(term, to_units=getattr(pyo_units, convert_to))
            
        return term_new
    
    def check_expression_units(self, convert_to=None):
        """Method to call in order to check an expressions units. This updates the expressions in place.

        :param str convert_to: The units to convert the expression to

        :return: None

        """
        scalar = 1
        if convert_to is None:
            raise ValueError('You need to supply a conversion')
        expr = self.expression
        expr_new = 0
        if isinstance(expr, NegationExpression):
            scalar *= -1
        if isinstance(self.expression, (DivisionExpression, ProductExpression)):
            term = self._check_term(expr, convert_to)
            expr_new = scalar*term
        else:
            if isinstance(expr, (int, float)):
                return None
            for i, term in enumerate(expr.args):
                term = self._check_term(term, convert_to)
                expr_new += scalar*term
        self.expression = expr_new
        self.units = getattr(pyo_units, convert_to)
             
        return None



"""
Diagnostic Tools used in Kipet

.. note::

    These methods should be checked for usefulness and possibly removed

"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy


def rank(A, eps=1e-10):
    """ obtains the rank of a matrix based on SVD

    :param numpy.ndarray A: A numpy matrix
    :param float eps: the value of the singular values that corresponds to 0
        when smaller than eps. Default = 1e-10

    :return rank: The rank of the matrix
    :rtype: int

    """
    print(type(A))  
    if isinstance(A, np.matrix):
        u, s, vh = np.linalg.svd(A)
        return len([x for x in s if abs(x) > eps]) 
    elif isinstance(A, pd.core.frame.DataFrame): 
        A = np.array(A)
        U, s, V = np.linalg.svd(A, full_matrices=True)
        return len([x for x in s if abs(x) > eps]) 
    else:
        raise RuntimeError("Must provide A as either numpy matrix or pandas dataframe")


def nullspace(A, atol=1e-13, rtol=0):
    """Obtains the nullspace of a matrix based on SVD. Taken from the SciPy cookbook

    If `A` is an array with shape (m, k), then `ns` will be an array
        with shape (k, n), where n is the estimated dimension of the
        nullspace of `A`.  The columns of `ns` are a basis for the
        nullspace; each element in numpy.dot(A, ns) will be approximately
        zero.

    :param numpy.ndarray A: A numpy matrix
    :param float atol: The absolute tolerance for a zero singular value.  Singular values
        smaller than `atol` are considered to be zero.
    :param float rtol: The relative tolerance.  Singular values less than rtol*smax are
        considered to be zero, where smax is the largest singular value.

    :return ns:
    :rtype: numpy.ndarray

    """
    A = np.atleast_2d(A)
    u, s, vh = scipy.linalg.svd(A)
    tol = max(atol, rtol * s[0])
    nnz = (s >= tol).sum()
    ns = vh[nnz:].conj().T
    return ns


def basic_pca(dataFrame, n=None, with_plots=False):
    """ Runs basic component analysis based on SVD

    :param pandas.DataFrame dataFrame: spectral data
    :param int n: number of largest singular-values to plot
    :param bool with_plots: Argument for files with plots due to testing

    :return: None

    """
            
    times = np.array(dataFrame.index)
    lambdas = np.array(dataFrame.columns)
    D = np.array(dataFrame)
    U, s, V = np.linalg.svd(D, full_matrices=True)

    if n is None:
        print("WARNING: since no number of components is specified, all components are printed")
        print("It is advised to select the number of components for n")
        n_shape = s.shape
        n = n_shape[0]
        
    u_shape = U.shape
    n_l_vector = n if u_shape[0] >= n else u_shape[0]
    n_singular = n if len(s) >= n else len(s)
    idxs = range(n_singular)
    vals = [s[i] for i in idxs]
    v_shape = V.shape
    n_r_vector = n if v_shape[0] >=n else v_shape[0]
    
    if with_plots:
        for i in range(n_l_vector):
            plt.plot(times, U[:, i])
        plt.xlabel("time")
        plt.ylabel("Components U[:,i]")
        plt.show()
        
        plt.semilogy(idxs, vals, 'o')
        plt.xlabel("i")
        plt.ylabel("singular values")
        plt.show()
        
        for i in range(n_r_vector):
            plt.plot(lambdas, V[i, :])
        plt.xlabel("wavelength")
        plt.ylabel("Components V[i,:]")
        plt.show()


def perform_data_analysis(dataFrame, pseudo_equiv_matrix, rank_data):  
    """ Runs the analysis by Chen, et al, 2018, based upon the pseudo-equivalency
    matrix. User provides the data and the pseudo-equivalency matrix and the analysis
    provides suggested number of absorbing components as well as whether there are
    likely to be unwanted spectral contributions.

    :param pandas.DataFrame dataFrame: Spectral data
    :param list pseudo_equiv_matrix: List containing the rows of the pseudo-equivalency matrix.
    :param int rank_data: Rank of the data matrix, as determined from SVD (number of coloured species)
    :param bool with_plots: Argument for files with plots due to testing

    :return: None

    """  
    if not isinstance(dataFrame, pd.DataFrame):
        raise TypeError("data must be inputted as a pandas DataFrame, try using read_spectral_data_from_txt or similar function first")
    
    if not isinstance(pseudo_equiv_matrix, list):
        raise TypeError("The Pseudo-equivalency matrix must be inputted as a list containing lists with each row of the pseudo-equivalency matrix")
    PEM = np.matrix(pseudo_equiv_matrix)
    rkp = rank(PEM)
    print("Rank of pseudo-equivalency matrix is ", rkp)
    
    ns = nullspace(PEM)
    print("Nullspace/kernel of pseudo-equivalency matrix is ", ns)
    if ns.size == 0:
        print("Null space of pseudo-equivalency matrix is null")
        rankns = 0
    else:
        rankns = rank(ns)
    
    print("the rank of the nullspace/kernel of pseudo-equivalency matrix is ", rankns)
    
    num_components = PEM.shape[1]
    if rankns > 0:
        ncr = num_components - rankns
        print("Choose the following number of absorbing species:", ncr)
    else:
        ncr = num_components
    ncms = rank_data
    
    if ncr == ncms:
        print("Solve standard problem assuming no unwanted contributions")
    elif ncr == ncms - 1:
        print("Solve with unwanted contributions")
    else:
        print("There may be uncounted for species in the model, or multiple sources of unknown contributions")



"""
KipetModel Mixins
"""
# Standard library imports
import copy
import os
import pathlib
import sys
import time

# Kipet library imports
from kipet.estimator_tools.parameter_estimator import wavelength_subset_selection

# Thirdparty library imports 
import plotly.graph_objects as go
import plotly.io as pio


class WavelengthSelectionMixins:
    
    """Wrapper class mixin of wavelength subset selection methods for ReactionModel"""
    
    def lack_of_fit(self):
        """Wrapper for ParameterEstimator lack_of_fit method

        :return: lof from p_estimator

        """
        lof = self.p_estimator.lack_of_fit()
        return lof
        
    def wavelength_correlation(self, corr_plot=False):
        """Wrapper for wavelength_correlation method in ParameterEstimator
        
        :param bool corr_plot: Option to plot the correlation vs wavelength plot
        
        :return pandas.DataFrame correlations: The correlation data
        
        """
        correlations = self.p_estimator.wavelength_correlation()
    
        if corr_plot:
            
            plot_options = {
                'label_font': dict(
                        size=32,
                    ),
                'title_font': dict(
                        size=32,
                    ),
                'tick_font': dict(
                        size=24,
                    ),
                }
            
            filename = "wavelength-correlations.html"
            fig = go.Figure()
            lists1 = sorted(correlations.items())
            x1, y1 = zip(*lists1)
            
            line = dict(width=4)
            fig.add_trace(
                go.Scatter(x=x1,
                           y=y1,
                           line=line,
                   )
                )
            
            fig.update_layout(
                title="Wavelength Correlations",
                xaxis_title="Wavelength (cm)",
                yaxis_title="Correlation",
                )
            plot_method = pio.write_html 
            
            t = time.localtime()
            date = f'{t.tm_year}-{t.tm_mon:02}-{t.tm_mday:02}-{t.tm_hour:02}-{t.tm_min:02}-{t.tm_sec:02}'
            
            fig.update_xaxes(zeroline=True, zerolinewidth=2, zerolinecolor='#4e4e4e', titlefont=plot_options['label_font'], tickfont=plot_options['tick_font'])
            fig.update_yaxes(zeroline=True, zerolinewidth=2, zerolinecolor='#4e4e4e', titlefont=plot_options['label_font'], tickfont=plot_options['tick_font'])
            
            fig.update_layout(title_font=plot_options['title_font'])
            fig.update_layout(legend_font=plot_options['label_font'])
            
            # if self.jupyter:
            #     chart_dir = Path.cwd().joinpath('charts', f'{self.name}-{folder_name}')
            #     plot_method = pio.show
            #     fig.update_layout(         
            #         autosize=False,
            #         width=1200,
            #         height=800,
            #         margin=dict(
            #             l=50,
            #             r=50,
            #             b=50,
            #             t=50,
            #             pad=4
            #         ),
            #     )
            # else:
            #     calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
            #     chart_dir = Path(calling_file_name).joinpath('charts', f'{self.name}-{folder_name}')
            #     plot_method = pio.write_html

            calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
            chart_dir = pathlib.Path(calling_file_name).joinpath('charts', f'{self.name}-{date}')
            
            chart_dir.mkdir(parents=True, exist_ok=True)
            filename = chart_dir.joinpath(filename)
            print(f'Plot saved as: {filename}')
            
            self.save_static_image = True
            if self.save_static_image:
                fig.write_image(f'{filename}.svg', width=1400, height=900)

            plot_method(fig, file=filename.as_posix(), auto_open=False)

        self.wavelength_correlations = correlations
        return correlations
    
    def run_lof_analysis(self, **kwargs):
        """Wrapper for run_lof_analysis method in ParameterEstimator

        :param dict kwargs: The options to pass through to the ParameterEstimator object

        :return: None

        """
        builder_before_data = copy.copy(self._builder)
        builder_before_data.clear_data()
        
        end_time = self._builder.end_time
        
        correlations = self.wavelength_correlation(corr_plot=False)
        lof = self.lack_of_fit()
        
        nfe = self.settings.collocation.nfe
        ncp = self.settings.collocation.ncp
        sigmas = self.settings.parameter_estimator.variances
        
        self.p_estimator.run_lof_analysis(builder_before_data, end_time, correlations, lof, nfe, ncp, sigmas, **kwargs)

        return None
    
    def wavelength_subset_selection(self, n=0):
        """Wrapper for wavelength_subset_selection method in ParameterEstimator

        :param float n: The subset of wavelengths to select (based on correlation)

        :return: The results of the wavelength_subset_selection method in ParameterEstimator

        """
        if n == 0:
            raise ValueError('You need to choose a subset!')
            
        if not hasattr(self, 'wavelength_correlations'):
            correlations = self.wavelength_correlation(corr_plot=False)
        else:
            correlations = self.wavelength_correlations
        
        return wavelength_subset_selection(correlations=correlations, n=n)
    
    def run_opt_with_subset_lambdas(self, wavelength_subset, **kwargs):
        """Wrapper for run_param_est_with_subset_lambdas method in ParameterEstimator

        :param list wavelength_subset: The subset of chosen wavelengths
        :param dict kwargs: The dict of options to pass through (here the solver)

        :return: Parameter estimation results with the given subset
        :rtype: ResultsObject
        """
        solver = kwargs.pop('solver', 'k_aug')    
    
        builder_before_data = copy.copy(self._builder)
        builder_before_data.clear_data()
        end_time = self.p_estimator.model.end_time.value
        nfe = self.settings.collocation.nfe
        ncp = self.settings.collocation.ncp
        sigmas = self.settings.parameter_estimator.variances
        
        results = self.p_estimator.run_param_est_with_subset_lambdas(builder_before_data, end_time, wavelength_subset, nfe, ncp, sigmas, solver=solver)    
        return results



"""
Additional Input Options for Variance and Parameter Estimation Classes
"""


def add_inputs(est_object, kwds):
    """This method takes an estimator object and some keyword arguments. The keyword arguments are then placed
    into the correct variables or methods.

    All changes are made in place. This used to be inside the estimator classes, but is cleaner if left here.

    :param est_object: This is a VarianceEstimator or ParameterEstimator instance
    :param dict kwds: A dict of keyword arguments

    :return: None

    """
    est_object.fixedtraj = kwds.pop("fixedtraj", False)
    est_object.fixedy = kwds.pop("fixedy", False)
    est_object.inputs_sub = kwds.pop("inputs_sub", None)
    est_object.yfix = kwds.pop("yfix", None)
    est_object.yfixtraj = kwds.pop("yfixtraj", None)
    trajectories = kwds.pop("trajectories", None)

    if est_object.inputs_sub is not None:
        for k in est_object.inputs_sub.keys():
            if not isinstance(est_object.inputs_sub[k], list):
                print("wrong type for inputs_sub {}".format(type(est_object.inputs_sub[k])))
                # raise Exception
            for i in est_object.inputs_sub[k]:
                # print(est_object.inputs_sub[k])
                # print(i)
                if est_object.fixedtraj is True or est_object.fixedy is True:
                    if est_object.fixedtraj is True:
                        for j in est_object.yfixtraj.keys():
                            for l in est_object.yfixtraj[j]:
                                if i == l:
                                    # print('herel:fixedy', l)
                                    if not isinstance(est_object.yfixtraj[j], list):
                                        print("wrong type for yfixtraj {}".format(type(est_object.yfixtraj[j])))
                                    reft = trajectories[(k, i)]
                                    est_object.fix_from_trajectory(k, i, reft)
                    if est_object.fixedy is True:
                        for j in est_object.yfix.keys():
                            for l in est_object.yfix[j]:
                                if i == l:
                                    if not isinstance(est_object.yfix[j], list):
                                        print("wrong type for yfix {}".format(type(est_object.yfix[j])))
                                    for key in est_object.model.alltime.value:
                                        vark = getattr(est_object.model, k)
                                        vark[key, i].set_value(key)
                                        vark[key, i].fix()  # since these are inputs we need to fix this
                else:
                    print("A trajectory or fixed input is missing for {}\n".format((k, i)))

    return None



# Standard library imports
from os import getcwd

# Third party imports
from pyomo.dae import ContinuousSet
from pyomo.environ import Constraint, ConstraintList, Param, TransformationFactory, value, Var
from pyomo.opt import ProblemFormat, SolverFactory, TerminationCondition

# KIPET library imports
from kipet.model_tools.visitor_classes import ReplacementVisitor
from kipet.general_settings.variable_names import VariableNames
from kipet.model_tools.pyomo_model_tools import (change_continuous_set,
                                                 get_index_sets)

__author__ = 'David M Thierry, Kevin McBride'  #: April 2018 - May 2021


class FEInitialize(object):
    """This class implements the finite per finite element initialization for
    a pyomo model initialization. A march-forward simulation will be run and 
    the resulting data will be patched to the tgt_model.
    
    The current strategy is as follows:
    1. Create a copy of the undiscretized model.
    2. Change the corresponding time set bounds to (0,1).
    3. Discretize and create the new model with the parameter h_i.
    4. Deactivate initial conditions.
    5. Check for params and inputs.

    
    .. note::
        
        An input needs to be a variable(fixed) indexed over time. Otherwise 
        it would be a parameter.

    """

    def __init__(self,
                 model_orig,
                 src_mod,
                 init_con=None,
                 param_name=None,
                 param_values=None,
                 inputs=None,
                 inputs_sub=None,
                 jump_times=None,
                 jump_states=None,
                 ):
        """
        The `the paran name` might be a list of strings or a single string
          corresponding to the parameters of the model.
        The `param_values` dictionary needs to be declared with the following 
        syntax: `param_dict["P", "k0"] = 49.7796`
        
        Where the first key corresponds to one of the parameter names, and the
        second to the corresponding index (if any).
        
        A similar structure is expected for the initial conditions and inputs.

        The `inputs` and `input_sub` parameters are in place depending of 
        whether there is a single index input or a multiple index input.

        Note that if the user does not provide correct information to 
        fe_factory; an exception will be thrown because of the n_var and m_eqn
        check for simulation.

        Once the constructor is called, one can initialize the model with the 
        following sintax: `self.load_initial_conditions(init_cond=ics_dict)`

        Finally, to run the initialization and automatic data patching to tgt
        model use: `self.run()`

        If a given finite element problem fails, we do will try once again with
        relaxed options. It is recommended to go back and check the model for 
        better understanding of the issue.

        Finally, an explicit function of time on the right hand side is 
        prohibited. Please put this information into an input (fixed variable)
        instead.

        :param ConcreteModel tgt_mod: The original fully discretized model that we want to patch the information to.
        :param ConcreteModel src_mod: The undiscretized reference model.
        :param str init_con: The initial constraint name (corresponds to a Constraint object).
        :param list param_name: The param name list. (Each element must correspond to a pyomo Var)
        :param dict param_values: The corresponding values: `param_dict["param_name", "param_index"] = 49.7796`
        :param dict inputs: The input dictionary. Use this dictonary for single index (time) inputs
        :param dict inputs_sub: The multi-index dictionary. Use this dictionary for multi-index inputs.
        
        """
        self.__var = VariableNames()

        self.ip = SolverFactory('ipopt')
        self.ip.options['halt_on_ampl_error'] = 'yes'
        self.ip.options['print_user_options'] = 'yes'

        self.model_orig = model_orig
        self.model_ref = src_mod.clone()

        self.volume_name = self.__var.volume_name
        if self.volume_name is None:
            raise ValueError('A volume name must exist')

        time_index = None
        for i in self.model_ref.component_objects(ContinuousSet):
            time_index = i
            break
        if time_index is None:
            raise Exception('no continuous_set')

        self.time_set = time_index.name

        model_original_time_set = getattr(self.model_orig, self.time_set)
        self.ncp = model_original_time_set.get_discretization_info()['ncp']
        fe_l = model_original_time_set.get_finite_elements()

        self.fe_list = [fe_l[i + 1] - fe_l[i] for i in range(0, len(fe_l) - 1)]
        self.nfe = len(self.fe_list)

        #: Re-construct the model with [0,1] time domain
        times = getattr(self.model_ref, self.time_set)
        change_continuous_set(times, [0, 1])

        for var in self.model_ref.component_objects(Var):
            var.clear()
            var.reconstruct()

        for con in self.model_ref.component_objects(Constraint):
            con.clear()
            con.construct()

        # self.model_ref.display(filename="selfmoddisc0.txt")
        #: Discretize
        d = TransformationFactory('dae.collocation')
        d.apply_to(self.model_ref, nfe=1, ncp=self.ncp, scheme='LAGRANGE-RADAU')

        #: Find out the differential variables
        self.dvs_names = []
        self.dvar_names = []

        for con in self.model_ref.component_objects(Constraint):
            name = con.name
            namel = name.split('_', 1)
            if len(namel) > 1:
                if namel[1] == "disc_eq":
                    realname = getattr(self.model_ref, namel[0])
                    self.dvar_names.append(namel[0])
                    self.dvs_names.append(realname.get_state_var().name)
        self.model_ref.h_i = Param(times, mutable=True, default=1.0)  #: Length of finite element

        #: Modify the collocation equations to introduce h_i (the length of finite element)
        for i in self.dvar_names:
            con = getattr(self.model_ref, i + '_disc_eq')
            dv = getattr(self.model_ref, i)
            e_dict = {}
            fun_tup = True
            for k in con.keys():
                if isinstance(k, tuple):
                    pass
                else:
                    k = (k,)
                    fun_tup = False
                e = con[k].expr.args[0]
                e_dict[k] = e * self.model_ref.h_i[k[0]] + dv[k] * (
                        1 - self.model_ref.h_i[k[0]]) == 0.0  #: As long as you don't clone
            if fun_tup:
                self.model_ref.add_component(i + "_deq_aug",
                                             Constraint(con.index_set(),
                                                        rule=lambda m, *j: e_dict[j] if j[
                                                                                            0] > 0.0 else Constraint.Skip))
            else:
                self.model_ref.add_component(i + "_deq_aug",
                                             Constraint(con.index_set(),
                                                        rule=lambda m, j: e_dict[j] if j > 0.0 else Constraint.Skip))
            self.model_ref.del_component(con)

        #: Sets for iteration
        #: Differential variables
        self.remaining_set = {}
        for i in self.dvs_names:
            dv = getattr(self.model_ref, i)
            if dv.index_set().name == times.name:  #: Just time set
                # print(i, 'here')
                self.remaining_set[i] = None
                continue
            # set_i = dv._implicit_subsets  #: More than just time set
            # set_i = dv._index._implicit_subsets #Update for pyomo 5.6.8 KH.L
            set_i = identify_member_sets(dv)
            # set_i = identify_member_sets(dv.index_set())
            # print(f'set_i = {set_i}')
            remaining_set = set_i[1]
            for s in set_i[2:]:
                remaining_set *= s
            if isinstance(remaining_set, list):
                self.remaining_set[i] = remaining_set
            else:
                self.remaining_set[i] = []
                self.remaining_set[i].append(remaining_set)
        #: Algebraic variables
        self.weird_vars = []  #:Not indexed by time
        self.remaining_set_alg = {}
        # with open('model_check.txt', 'w') as f5:
        #     self.model_ref.pprint(ostream = f5)
        for av in self.model_ref.component_objects(Var):

            # print(av.name)
            if av.name in self.dvs_names:
                continue
            if av.index_set().name == times.name:  #: Just time set
                self.remaining_set_alg[av.name] = None
                continue
            # set_i = av._implicit_subsets
            # set_i = av._index._implicit_subsets #Update for pyomo 5.6.8 KH.L
            set_i = identify_member_sets(av)
            if set_i is None or not times in set_i:
                self.weird_vars.append(av.name)  #: Not indexed by time!
                continue  #: if this happens we might be in trouble
            remaining_set = set_i[1]  #: Index by time and others
            for s in set_i[2:]:
                if s.name == times.name:
                    self.remaining_set_alg[av.name] = None
                    continue
                else:
                    remaining_set *= s
            if isinstance(remaining_set, list):
                self.remaining_set_alg[av.name] = remaining_set
            else:
                self.remaining_set_alg[av.name] = []
                self.remaining_set_alg[av.name].append(remaining_set)

        if init_con is not None:  #: Delete the initial conditions (we use .fix() instead)
            ic = getattr(self.model_ref, init_con)
            self.model_ref.del_component(ic)

        if isinstance(param_name, list):  #: Time independent parameters
            if param_values:
                if isinstance(param_values, dict):
                    for pname in param_name:
                        p = getattr(self.model_ref, pname)
                        for key in p.keys():
                            try:
                                val = param_values[pname, key]
                                p[key].set_value(val)
                            except KeyError:
                                raise Exception("Missing a key of the param_values\n"
                                                "Please provide all the required keys.\n"
                                                "missing: {}".format(key))
                            p[key].fix()
                else:
                    Exception("Arg param_values should be provided in a dictionary")
            else:
                Exception("Arg param_values should be provided in a dictionary")
        elif isinstance(param_name, str):
            if param_values:
                if isinstance(param_values, dict):
                    p = getattr(self.model_ref, param_name)
                    for key in p.keys():
                        try:
                            val = param_values[param_name, key]
                            p[key].set_value(val)
                        except KeyError:
                            raise Exception("Missing a key of the param_values\n"
                                            "Please provide all the required keys.\n"
                                            "missing: {}".format(key))
                        p[key].fix()
        elif not param_name:
            pass
        else:
            raise Exception("wrong type for param_name")

        #: Fix initial conditions
        for i in self.dvs_names:
            dv = getattr(self.model_ref, i)
            if self.remaining_set[i] is None:
                dv[0].fix()
            for rs in self.remaining_set[i]:
                for k in rs:
                    k = k if isinstance(k, tuple) else (k,)
                    dv[(0,) + k].fix()

        self.inputs = None
        self.input_remaining_set = {}

        #: Check if inputs are declared
        # if self.inputs is not None:
        #     if not isinstance(inputs, dict) or isinstance(inputs, str):
        #         raise Exception("Must be a dict or str")
        #     if isinstance(inputs, str):
        #         self.inputs = [self.inputs]
        #     for i in self.inputs:
        #         p = getattr(self.model_ref, i)
        #         p.fix()
        #         if p.index_set().name == times.name:  #: Only time-set
        #             self.input_remaining_set[i] = None
        #             continue
        #         #set_i = p._implicit_subsets
        #         #set_i = p._index._implicit_subsets #Update for pyomo 5.6.8 KH.L
        #         set_i = identify_member_sets(p)
        #         if not times in set_i:
        #             raise RuntimeError("{} is not by index by time, this can't be an input".format(i))
        #         remaining_set = set_i[1]
        #         for s in set_i[2:]:
        #             if s.name == times.name:  #: would this ever happen?
        #                 continue
        #             else:
        #                 remaining_set *= s
        #         if isinstance(remaining_set, list):
        #             self.input_remaining_set[i] = remaining_set
        #         else:
        #             self.input_remaining_set[i] = []
        #             self.input_remaining_set[i].append(remaining_set)

        # self.inputs_sub = None
        # inputs_sub['some_var'] = ['index0', 'index1', ('index2a', 'index2b')]

        self.inputs_sub = inputs_sub
        # print([i.name for i in get_index_sets(getattr(self.model_ref, 'Dose'))])

        if self.inputs_sub is not None:
            for key in self.inputs_sub.keys():
                model_var_obj = getattr(self.model_ref, key)

                # This finds the index of the set    
                # if identify_member_sets(model_var_obj) is None:
                #     raise RuntimeError("This variable is does not have multiple indices"
                #                         "Pass {} as part of the inputs keyarg instead.".format(key))
                # elif model_var_obj.index_set().name == times.name:
                #     raise RuntimeError("This variable is indexed over time"
                #                         "Pass {} as part of the inputs keyarg instead.".format(key))
                # else:
                #     if times not in identify_member_sets(model_var_obj):
                #         raise RuntimeError("{} is not indexed over time; it can not be an input".format(key))

                for k in self.inputs_sub[key]:
                    if isinstance(k, str) or isinstance(k, int) or isinstance(k, tuple):
                        k = (k,) if not isinstance(k, tuple) else k
                    else:
                        raise RuntimeError("{} is not a valid index".format(k))

                    for t in times:
                        model_var_obj[(t,) + k].fix()

        if hasattr(self.model_ref, self.__var.time_step_change):
            for time_step in getattr(self.model_ref, self.__var.time_step_change):
                getattr(self.model_ref, self.__var.time_step_change)[time_step].fix()

        if hasattr(self.model_ref, self.__var.model_constant):
            for param, obj in getattr(self.model_ref, self.__var.model_constant).items():
                obj.fix()

        # if hasattr(self.model_ref, self.__var.algebraic):
        #     model_var_obj = getattr(self.model_ref, self.__var.algebraic)
        #     for k in ['f']:
        #         if isinstance(k, str) or isinstance(k, int) or isinstance(k, tuple):
        #             k = (k,) if not isinstance(k, tuple) else k
        #         else:
        #             raise RuntimeError("{} is not a valid index".format(k))
        #         for t in times:
        #             print(model_var_obj.display())
        #             model_var_obj[(t,) + k].fix()

        # for param, obj in getattr(self.model_ref, self.__var.algebraic).items():
        #     print(param)
        #     print(obj)
        #     if param[1] == 'f' or param[1] == 'Csat':
        #         for t in times:
        #             print(t, param[1])
        #             obj[(t,) + (param[1],)].fix()

        # : Check n vars and m equations
        (n, m) = reconcile_nvars_mequations(self.model_ref)
        if n != m:
            raise Exception("Inconsistent problem; n={}, m={}".format(n, m))
        self.jump = False
        self.con_num = 0

    def load_initial_conditions(self, init_cond=None):
        if not isinstance(init_cond, dict):
            raise Exception("init_cond must be a dictionary")

        for i in self.dvs_names:
            dv = getattr(self.model_ref, i)
            ts = getattr(self.model_ref, self.time_set)  # self.model_ref.alltime
            for t in ts:
                for s in self.remaining_set[i]:
                    if s is None:
                        val = init_cond[i]  #: if you do not have an extra index, just put the value there
                        dv[t].set_value(val)
                        if t == 0:
                            if not dv[0].fixed:
                                dv[0].fix()
                        continue
                    for k in s:
                        val = init_cond[i, k]
                        k = k if isinstance(k, tuple) else (k,)
                        dv[(t,) + k].set_value(val)
                        if t == 0:
                            if not dv[(0,) + k].fixed:
                                dv[(0,) + k].fix()

    def march_forward(self, fe):
        """Moves forward with the simulation.

        This method performs the actions required for setting up the `fe-th` problem.

        Adjust inputs.
        Solve current problem.
        Patches tgt_model.
        Cycles initial conditions

        :param int fe: The corresponding finite element.

        :return: None

        """
        self.adjust_h(fe)
        if self.inputs or self.inputs_sub:
            self.load_input(fe)

        self.ip.options["print_level"] = 1  #: change this on demand
        # self.ip.options["start_with_resto"] = 'no'
        self.ip.options['bound_push'] = 1e-02
        sol = self.ip.solve(self.model_ref, tee=False, symbolic_solver_labels=True)

        # Try to redo it if it fails
        if sol.solver.termination_condition != TerminationCondition.optimal:
            self.ip.options["OF_start_with_resto"] = 'yes'

            sol = self.ip.solve(self.model_ref, tee=False, symbolic_solver_labels=True)
            if sol.solver.termination_condition != TerminationCondition.optimal:

                self.ip.options["OF_start_with_resto"] = 'no'
                self.ip.options["bound_push"] = 1E-02
                self.ip.options["OF_bound_relax_factor"] = 1E-05
                sol = self.ip.solve(self.model_ref, tee=True, symbolic_solver_labels=True)
                self.ip.options["OF_bound_relax_factor"] = 1E-08

                # It if fails twice, raise an error
                if sol.solver.termination_condition != TerminationCondition.optimal:
                    raise Exception("The current iteration was unsuccessful. Iteration :{}".format(fe))

        self.patch(fe)
        self.cycle_ics(fe)

    def load_discrete_jump(self, dosing_points):
        """Method is used to define and load the places where discrete jumps are located, e.g.
        dosing points or external inputs.

        :param list dosing_points: A list of DosingPoint objects

        :return: None

        """
        self.jump = True
        self.dosing_points = dosing_points

        return None

    def cycle_ics(self, curr_fe):
        """Cycles the initial conditions of the initializing model.
        Take the values of states (initializing model) at t=last and patch them into t=0.
        Check: :ref:`<https://github.com/dthierry/cappresse/blob/pyomodae-david/nmpc_mhe/aux/utils.py>`_
        fe_cp function!

        :param int curr_fe: The current finite element

        :return: None

        """
        ts = getattr(self.model_ref, self.time_set)
        t_last = t_ij(ts, 0, self.ncp)

        # Inclusion of discrete jumps: (CS)
        for i in self.dvs_names:
            dv = getattr(self.model_ref, i)
            for s in self.remaining_set[i]:
                if s is None:
                    val = value(dv[t_last])
                    dv[0].set_value(val)
                    if not dv[0].fixed:
                        dv[0].fix()
                    continue
                for k in s:
                    k = k if isinstance(k, tuple) else (k,)
                    val = value(dv[(t_last,) + k])
                    dv[(0,) + k].set_value(val)
                    if not dv[(0,) + k].fixed:
                        dv[(0,) + k].fix()

    def patch(self, fe):
        """ Take the current state of variables of the initializing model at fe and load it into the tgt_model
        Note that this will skip fixed variables as a safeguard.

        :param int fe: The current finite element to be patched (tgt_model).

        :return: None

        """
        time_set_ref = getattr(self.model_ref, self.time_set)
        time_set_orig = getattr(self.model_orig, self.time_set)

        for model_ref_var in self.model_ref.component_objects(Var, active=True):
            model_orig_var = getattr(self.model_orig, model_ref_var.name)
            if model_ref_var.name in self.weird_vars:
                for k in model_ref_var.keys():
                    if model_ref_var[k].stale or model_ref_var[k].is_fixed():
                        continue
                    try:
                        val = model_ref_var[k].value
                    except ValueError:
                        pass
                    model_ref_var[k].set_value(val)
                continue
            #: From this point on all variables are indexed over time.
            if model_ref_var.name in self.dvs_names:
                drs = self.remaining_set[model_ref_var.name]
            else:
                drs = self.remaining_set_alg[model_ref_var.name]

            for j in range(0, self.ncp + 1):

                t_tgt = t_ij(time_set_orig, fe, j)
                t_src = t_ij(time_set_ref, 0, j)

                if drs is None:
                    if model_ref_var[t_src].stale or model_ref_var[t_src].is_fixed():
                        continue
                    try:
                        val = model_ref_var[t_src].value
                    except ValueError:
                        print("Error at {}, {}".format(model_ref_var.name, t_src))
                    model_ref_var[t_tgt].set_value(val)
                    continue

                for k in drs:
                    for key in k:
                        key = key if isinstance(key, tuple) else (key,)
                        if model_ref_var[(t_src,) + key].stale or model_ref_var[(t_src,) + key].is_fixed():
                            continue
                        try:
                            val = value(model_ref_var[(t_src,) + key])
                        except ValueError:
                            print("Error at {}, {}".format(model_ref_var.name, (t_src,) + key))
                        model_orig_var[(t_tgt,) + key].set_value(val)

        if self.jump:

            """This creates a new constraint forcing the variable at the
            specific time to be equal to step size provided in the dosing
            points. It creates the constraint and replaces the variable in the
            original ode equations.
            """
            vs = ReplacementVisitor()
            for model_var, dosing_point_list in self.dosing_points.items():

                for dosing_point in dosing_point_list:

                    self.jump_fe, self.jump_cp = fe_cp(time_set_orig, dosing_point.time)
                    comp_dict = self.make_comp_list()

                    if fe == self.jump_fe + 1:

                        for comp_tuple in comp_dict:
                            model_var = comp_tuple[0]
                            comp = comp_tuple[1]
                            con_name = f'd{model_var}dt_disc_eq'
                            varname = f'{model_var}_dummy_{self.con_num}'
                            conc_delta = self.concentration_calc(dosing_point)

                            # This is the constraint you want to change (add dummy for the model_var)
                            model_con_obj = getattr(self.model_orig, con_name)

                            # Adding some kind of constraint to the list
                            self.model_orig.add_component(f'{model_var}_dummy_eq_{self.con_num}_{comp}',
                                                          ConstraintList())
                            model_con_objlist = getattr(self.model_orig, f'{model_var}_dummy_eq_{self.con_num}_{comp}')

                            # Adding a variable (no set) with dummy name to model
                            self.model_orig.add_component(varname, Var([0]))

                            # vdummy is the var_obj of the Var you just made
                            vdummy = getattr(self.model_orig, varname)

                            # this is the variable that will replace the other
                            vs.change_replacement(vdummy[0])

                            # adding a parameter that is the jump at dosing_point.step (the size or change in the var)
                            self.model_orig.add_component(f'{model_var}_jumpdelta{self.con_num}_{comp}',
                                                          Param(initialize=conc_delta[comp]))

                            # This is the param you just made
                            jump_param = getattr(self.model_orig, f'{model_var}_jumpdelta{self.con_num}_{comp}')

                            # This is where the new concentrations need to be calculated - start with A
                            # jump_param is what needs to be modified

                            # Constraint setting the variable equal to the step size
                            exprjump = vdummy[0] - getattr(self.model_orig, model_var)[
                                (dosing_point.time,) + (comp,)] == jump_param

                            # Add the new constraint to the original model
                            self.model_orig.add_component(f'jumpdelta_expr{self.con_num}_{comp}',
                                                          Constraint(expr=exprjump))

                            for kcp in range(1, self.ncp + 1):
                                curr_time = t_ij(time_set_orig, self.jump_fe + 1, kcp)
                                idx = (curr_time,) + (comp,)
                                model_con_obj[idx].deactivate()
                                var_expr = model_con_obj[idx].expr
                                suspect_var = var_expr.args[0].args[1].args[0].args[0].args[1]
                                vs.change_suspect(id(suspect_var))  #: who to replace
                                e_new = vs.dfs_postorder_stack(var_expr)  #: replace
                                model_con_obj[idx].set_value(e_new)
                                model_con_objlist.add(model_con_obj[idx].expr)

                            self.con_num += 1

    def make_comp_list(self):
        """Creates a list of tuples to pair model variables with component
        and volume variables
        
        :return list comp_dict: A list of tuples
        
        """
        comp_dict = []
        for comp in self.model_orig.mixture_components.keys():
            comp_dict.append((self.__var.concentration_model, comp))
        comp_dict.append((self.__var.state_model, self.volume_name))
        return comp_dict

    def concentration_calc(self, dosing_point):
        """This method calculates the changes in the concentration when 
        adding a volume of substance with specified concentrations of species
        
        :param DosingPoint dosing_point: Takes a dosing point object
        
        :return dict delta_conc: A dict of tuples with the component and 
          step change in concentration. This includes the volume step too.
        
        .. note::
            
            This only takes a single species at the moment. Use multiple dosing
            points if you need to add mixtures at the same point.
            
        """
        time_set_orig = getattr(self.model_orig, self.time_set)

        # Get the current time point
        curr_time = t_ij(time_set_orig, self.jump_fe + 1, 1)

        # Get the current volume
        vol = getattr(self.model_orig, self.__var.state_model)[(curr_time,) + (self.volume_name,)].value

        # Get the current concentrations
        conc = {}
        for comp in list(self.model_orig.mixture_components.keys()):
            conc[comp] = getattr(self.model_orig, self.__var.concentration_model)[(curr_time,) + (comp,)].value

        # Calculate the moles of each substance at the current point
        moles = {}
        for comp in conc:
            moles[comp] = vol * conc[comp]

        conc_change = dosing_point.conc[0]
        vol_change = dosing_point.vol[0]

        # Add the dosing point to the moles
        moles[dosing_point.component] += conc_change * vol_change
        vol += vol_change
        delta_conc = {k: v / vol - conc[k] for k, v in moles.items()}
        delta_conc[self.volume_name] = vol_change

        return delta_conc

    def adjust_h(self, fe):
        """Adjust the h_i parameter of the initializing model.

        The initializing model goes from t=(0,1) so it needs to be scaled by the current time-step size.

        :param int fe: The current value of h_i

        :return: None

        """
        hi = getattr(self.model_ref, "h_i")
        zeit = getattr(self.model_ref, self.time_set)
        for t in zeit:
            hi[t].value = self.fe_list[fe]

    def run(self, resto_strategy="bound_relax"):
        """Runs the sequence of problems fe=0,nfe

        :param str resto_strategy: The restoration strategy for the march_forward algorithm

        :return: None

        """
        for i in range(0, len(self.fe_list)):
            self.march_forward(i)

    def load_input(self, fe):
        """ Loads the current value of input from tgt_model into the initializing model at the current fe.

        :param int fe:  The current finite element to be loaded.

        :return: None

        """
        if self.inputs is not None:
            time_set_ref = getattr(self.model_ref, self.time_set)
            time_set_orig = getattr(self.model_orig, self.time_set)
            for i in self.inputs:
                p_data = getattr(self.model_orig, i)
                p_sim = getattr(self.model_ref, i)
                if self.input_remaining_set[i] is None:
                    for j in range(0, self.ncp + 1):
                        t = t_ij(time_set_orig, fe, j)
                        tsim = t_ij(time_set_ref, 0, j)
                        val = value(p_data[t])
                        p_sim[tsim].set_value(val)
                    continue
                for k in self.input_remaining_set[i]:
                    for key in k:
                        for j in range(0, self.ncp + 1):
                            t = t_ij(time_set_orig, fe, j)
                            tsim = t_ij(time_set_ref, 0, j)
                            val = value(p_data[(t,) + key])
                            p_sim[(tsim,) + key].set_value(val)

        # Here is where the jumps come in... (can this be done with a different var?)
        if self.inputs_sub is not None:
            time_set_ref = getattr(self.model_ref, self.time_set)
            time_set_orig = getattr(self.model_orig, self.time_set)

            for key in self.inputs_sub.keys():  # Y
                model_orig_var = getattr(self.model_orig, key)
                model_ref_var = getattr(self.model_ref, key)

                for sub_key in self.inputs_sub[key]:
                    sub_key = (sub_key,) if not isinstance(sub_key, tuple) else k

                    for j in range(0, self.ncp + 1):
                        t_orig = t_ij(time_set_orig, fe, j)
                        t_ref = t_ij(time_set_ref, 0, j)
                        val = model_orig_var[(t_orig,) + sub_key].value
                        model_ref_var[(t_ref,) + sub_key].set_value(val)

    def create_bounds(self, bound_dict):
        time_set_ref = getattr(self.model_ref, self.time_set)
        for v in bound_dict.keys():
            var = getattr(self.model_ref, v)
            varbnd = bound_dict[v]
            if not isinstance(varbnd, dict):
                raise RuntimeError("The entry for {} is not a dictionary".format(v))
            for t in time_set_ref:
                for k in varbnd.keys():
                    bnd = varbnd[k]
                    if not isinstance(k, tuple):
                        k = (k,)
                    var[(t,) + k].setlb(bnd[0])  #: Lower bound
                    var[(t,) + k].setub(bnd[1])  #: Upper bound

    def clear_bounds(self):
        for v in self.model_ref.component_data_objects(Var):
            v.setlb(None)
            v.setub(None)


def t_ij(time_set, i, j):
    # type: (ContinuousSet, int, int) -> float
    """Return the corresponding time(continuous set) based on the i-th finite element and j-th collocation point
    From the NMPC_MHE framework by @dthierry.

    :param ContinuousSet time_set: Parent Continuous set
    :param int i: finite element
    :param int j: collocation point

    :return: Corresponding index of the ContinuousSet
    :rtype: float

    """
    if i < time_set.get_discretization_info()['nfe']:
        h = time_set.get_finite_elements()[i + 1] - time_set.get_finite_elements()[i]  #: This would work even for 1 fe
    else:
        h = time_set.get_finite_elements()[i] - time_set.get_finite_elements()[i - 1]  #: This would work even for 1 fe
    tau = time_set.get_discretization_info()['tau_points']
    fe = time_set.get_finite_elements()[i]
    time = fe + tau[j] * h
    return round(time, 6)


def write_nl(d_mod, filename=None):
    """
    Write the nl file

    :param ConcreteModel d_mod: the model of interest.

    :return str cwd: The current working directory.

    """
    if not filename:
        filename = d_mod.name + '.nl'
    d_mod.write(filename, format=ProblemFormat.nl,
                io_options={"symbolic_solver_labels": True})
    cwd = getcwd()
    # print("nl file {}".format(cwd + "/" + filename))
    return cwd


def reconcile_nvars_mequations(d_mod):
    """
    Compute the actual number of variables and equations in a model by reading the relevant line at the nl file.

    :param ConcreteModel d_mod: the model of interest.

    :return tuple: The number of variables and the number of constraints.

    """
    fullpth = getcwd()
    fullpth += "/_reconciled.nl"
    write_nl(d_mod, filename=fullpth)
    with open(fullpth, 'r') as nl:
        lines = nl.readlines()
        line = lines[1]
        newl = line.split()
        nvar = int(newl[0])
        meqn = int(newl[1])
        nl.close()

    return (nvar, meqn)


def disp_vars(model, file):
    """Helper function for debugging

    :param ConcreteModel model: the model of interest.
    :param str file: Destination text file.

    :return: None

    """
    with open(file, 'w') as f:
        for c in model.component_objects(Var):
            c.pprint(ostream=f)
        f.close()


def fe_cp(time_set, feedtime):
    """Return the corresponding fe and cp for a given time

    :param ContinuousSet time_set: The time index
    :param float feedtime: The time of the dosing point

    :return: tuple of the finite element and the collocation point

    """
    fe_l = time_set.get_lower_element_boundary(feedtime)
    fe = None
    j = 0
    for i in time_set.get_finite_elements():
        if fe_l == i:
            fe = j
            break
        j += 1
    h = time_set.get_finite_elements()[1] - time_set.get_finite_elements()[0]
    tauh = [i * h for i in time_set.get_discretization_info()['tau_points']]
    j = 0  #: Watch out for LEGENDRE
    cp = None
    for i in tauh:
        if round(i + fe_l, 6) == feedtime:
            cp = j
            break
        j += 1
    return fe, cp


def identify_member_sets(model_var_obj):
    """Identifies the index sets of the given variable

    :param GeneralVar model_var_obj: The variable object of the model

    :return: None or the list of indices

    """
    index_list = get_index_sets(model_var_obj)

    if len(index_list) > 1:
        return index_list
    else:
        return None



"""
Replacement functions

This module can be used to replace parameters in the models with constant
values followed by deletion of the parameters from the model.
"""
# KIPET library imports
from kipet.model_tools.visitor_classes import ReplacementVisitor


class ParameterReplacer:
    """This class replaces fixed parameters with numerical values

    :param list models: A list of models
    :param str parameter_name: The name of the parameter to change
    :param list expression_names: The expressions in which to change the parameter
    """

    def __init__(self,
                 models=None,
                 #fix_parameters=None,
                 parameter_name='P',
                 expression_names=['odes'],
                 #user_fixed=True,
                 ):

        """Initialization of ParameterReplacer

        Used to parse Pyomo expressions and replace parameters/variables with other values
        """

        self.models = models if models is not None else []
        #self.fix_parameters = fix_parameters
        self.parameter_name = parameter_name
        self.expression_names = expression_names
        #self.user_fixed = user_fixed

    def remove_fixed_vars(self):
        """Method to remove the fixed parameters from the ConcreteModel. This replaces the variables in place.

        :return: None
        """
        for model in self.models:
    
            # Change to use the direct models
            p_var = getattr(model, self.parameter_name)
    
            # Check which parameters are fixed and add them to the set:
            params_to_fix = set([k for k, v in p_var.items() if v.fixed])

            # Replace the parameters with numeric values and remove Var from model
            for param in params_to_fix:   
                parameter_to_change = param
                if parameter_to_change in p_var.keys():
                    change_value = p_var[param].value
                
                    for expression_var in self.expression_names:
                        e_var = getattr(model, expression_var)
        
                        for k, v in e_var.items(): 
                            ep_updated_expr = _update_expression(v.body, p_var[param], change_value)
                            e_var[k] = ep_updated_expr == 0
                
                    if hasattr(model, 'parameter_names'):
                        model.parameter_names.remove(param)
                    del p_var[param]
        
        return None


def _update_expression(expr, replacement_param, change_value):
    """Takes the non-estiambale parameter and replaces it with its initial value

    :param expression expr: The target ODE constraint in the Pyomo model
    :param str replacement_param: The non-estimable parameter to replace
    :param float change_value: The initial value for the above parameter

    :return expression new_expr: The updated constraints with the
        desired parameter replaced with a float
    
    """
    visitor = ReplacementVisitor()
    visitor.change_replacement(change_value)
    visitor.change_suspect(id(replacement_param))
    new_expr = visitor.dfs_postorder_stack(expr)
    
    return new_expr



"""
Parameter selection and update methods
"""
# Third party imports
import numpy as np
import pandas as pd


def parameter_ratios(model_object, reduced_hessian, Se, epsilon=1e-16):
    """This is Eq. 26 from Chen and Biegler 2020 where the ratio of the 
    standard deviation for each parameter is calculated.
    
    :param ConcreteModel model_object: The current Pyomo model
    :param numpy.ndarray reduced_hessian: The current reduced hessian
    :param list Se: The list of free parameters
    :param float epsilon: Tolerance parameter for small values

    :return: The ratio of predicted standard deviations to the parameter values and the eigenvalues
    :rtype: tuple

    """
    eigenvalues, eigenvectors = np.linalg.eigh(reduced_hessian)
    red_hess_inv = np.dot(np.dot(eigenvectors, np.diag(1.0 / abs(eigenvalues))), eigenvectors.T)
    d = red_hess_inv.diagonal()
    d_sqrt = np.asarray(np.sqrt(d)).ravel()
    rp = [d_sqrt[i] / max(epsilon, model_object.P[k].value) for i, k in enumerate(Se)]

    return rp, eigenvalues


def rank_parameters(model_object, reduced_hessian, param_list, epsilon=1e-16, eta=1e-1):
    """Performs the parameter ranking based using the Gauss-Jordan
    elimination procedure.
    
    :param ConcreteModel model_object: The current Pyomo model
    :param numpy.ndarray reduced_hessian: The current reduced hessian
    :param list param_list: The list of model parameters
    :param float epsilon: Tolerance parameter for small values
    :param float eta: Tolerance for variances
    
    :return list Se_update: The updated list of parameters in Se.
    :return list Sf_update: The updated list of parameters if Sf.
    :rtype: tuple

    """
    eigenvector_tolerance = 1e-15
    parameter_tolerance = 1e-12
    squared_term_1 = 0
    squared_term_2 = 0
    Sf_update = []
    Se_update = []
    M = {}

    param_set = set(param_list)
    param_elim = set()
    eigenvalues, U = np.linalg.eigh(reduced_hessian)

    df_eigs = pd.DataFrame(np.diag(eigenvalues),
                           index=param_list,
                           columns=[i for i, x in enumerate(param_list)])
    df_U_gj = pd.DataFrame(U,
                           index=param_list,
                           columns=[i for i, x in enumerate(param_list)])

    # Gauss-Jordan elimination
    for i, p in enumerate(param_list):
        piv_col = i
        piv_row = df_U_gj.loc[:, piv_col].abs().idxmax()
        piv = (piv_row, piv_col)
        rows = list(param_set.difference(param_elim))
        M[i] = piv_row
        df_U_gj = _gauss_jordan_step(df_U_gj, piv, rows)
        param_elim.add(piv_row)
        df_eigs.drop(index=[param_list[piv_col]], inplace=True)
        df_eigs.drop(columns=[piv_col], inplace=True)

    # Parameter ranking
    eigenvalues, eigenvectors = np.linalg.eigh(reduced_hessian)
    ranked_parameters = {k: M[abs(len(M) - 1 - k)] for k in M.keys()}

    for k, v in ranked_parameters.items():
        name = v.split('[')[-1].split(']')[0]
        squared_term_1 += abs(1 / max(eigenvalues[-(k + 1)], parameter_tolerance))
        squared_term_2 += (eta ** 2 * max(abs(model_object.P[name].value), epsilon) ** 2)

        if squared_term_1 >= squared_term_2:
            Sf_update.append(name)
        else:
            Se_update.append(name)

    if len(Se_update) == 0:
        Se_update.append(ranked_parameters[0])
        Sf_update.remove(ranked_parameters[0])

    return Se_update, Sf_update


def _gauss_jordan_step(df_U_update, pivot, rows):
    """Performs the Gauss-Jordan Elimination step in W. Chen's method
    
    :param pandas.DataFrame df_U_update: A pandas DataFrame instance of the U matrix from the
            eigenvalue decomposition of the reduced hessian (can be obtained
            through the HessianObject).
    :param str pivot: The element where to perform the elimination.
    :param set rows: A set containing the rows that have already been eliminated.
        
    :return: The df_U_update DataFrame is returned after one row is eliminated.
    :rtype: pandas.DataFrame

    """
    if isinstance(pivot, tuple):
        pivot = dict(r=pivot[0],
                     c=pivot[1]
                     )

    uij = df_U_update.loc[pivot['r'], pivot['c']]

    for col in df_U_update.columns:
        if col == pivot['c']:
            continue

        factor = df_U_update.loc[pivot['r'], col] / uij

        for row in rows:
            if row not in rows:
                continue
            else:
                df_U_update.loc[row, col] -= factor * df_U_update.loc[row, pivot['c']]

    df_U_update[abs(df_U_update) < 1e-15] = 0
    df_U_update.loc[pivot['r'], pivot['c']] = 1

    return df_U_update



"""
General scaling methods for Kipet models

.. warning::

        This class is still under development and may change in future versions!

"""
# Standard library imports
import copy

# Third party imports
from pyomo.environ import Param

from kipet.model_tools.visitor_classes import ReplacementVisitor, ScalingVisitor
# Kipet library imports
from kipet.general_settings.variable_names import VariableNames

__var = VariableNames()


def scale_models(models_input, k_vals, name=None):
    """Takes a model or dict of models and iterates through them to update the
    odes and parameter values

    :param array-like models_input: The models to be updated
    :param dict k_vals: The current parameter values
    :param str name: The name of the model

    :return: The dict of ? and the dict of models
    :rtype: tuple

    """
    models_dict = copy.deepcopy(models_input)
    
    if not isinstance(models_dict, dict):
        key = 'model-1' if name is None else name
        models_dict = {key: models_dict}
    
    d_init = {}
    
    for model in models_dict.values():
        d_init_model = _scale_model(model, k_vals)
        d_init.update(d_init_model)

    return d_init, models_dict


def _scale_model(model, k_vals):
    """Scales an individual model and updates the initial parameter dict

    :param ConcreteModel model: The Pyomo model to be scaled
    :param dict k_vals: The current parameter values

    :return: The scaled parameter dict
    :rtype: dict

    """
    scaled_bounds = {}    

    if not hasattr(model, __var.model_parameter_scaled):
        add_scaling_parameters(model, k_vals)
        scale_parameters(model)
     
        for k, v in getattr(model, __var.model_parameter).items():
            lb, ub = v.bounds
            
            lb = lb/getattr(model, __var.model_parameter_scaled)[k].value
            ub = ub/getattr(model, __var.model_parameter_scaled)[k].value
            
            if ub < 1:
                print('Bounds issue, pushing upper bound higher')
                ub = 1.1
            if lb > 1:
                print('Bounds issue, pushing lower bound lower')
                lb = 0.9
                
            scaled_bounds[k] = lb, ub
                
            model_params = getattr(model, __var.model_parameter)
            model_params[k].setlb(lb)
            model_params[k].setub(ub)
            model_params[k].unfix()
            model_params[k].set_value(1)
            
    parameter_dict = {k: (1, scaled_bounds[k]) for k in k_vals.keys() if k in model_params}
            
    return parameter_dict


def add_scaling_parameters(model, k_vals=None):
    """This will actually set up the model variables in the Pyomo models

    :param ConcreteModel model: A pyomo model
    :param dict k_vals: A dict of parameter values

    :return: None

    """
    print("The model has not been scaled and will now be scaled using K parameters")
    
    if k_vals is None:
    
        setattr(model, __var.model_parameter_scaled, Param(model.parameter_names,                                                                                                                                                            
                    initialize={k: v for k, v in getattr(model, __var.model_parameter).items()},
                    mutable=True,
                    default=1))
    else:
        setattr(model, __var.model_parameter_scaled, Param(model.parameter_names,                                                                                                                                                            
                    initialize={k: v for k, v in k_vals.items() if k in getattr(model, __var.model_parameter)},
                    mutable=True,
                    default=1))
    
    return None


def scale_parameters(model, k_vals=None):
    """If scaling, this multiplies the constants in model.K to each
    parameter in model.P.
    
    I am not sure if this is necessary and will look into its importance.
    """
    if not hasattr(model, __var.model_parameter_scaled):
        add_scaling_parameters(model, k_vals=k_vals)
        
    scale = {}

    for var in __var.model_vars + __var.rate_vars:
        if hasattr(model, var):
            for i in getattr(model, var):
                if var == __var.model_parameter:
                    scale[id(getattr(model, var)[i])] = getattr(model, __var.model_parameter_scaled)[i]
                    continue
                scale[id(getattr(model, var)[i])] = 1

    for k, v in getattr(model, __var.ode_constraints).items():
        scaled_expr = _scale_expression(v.body, scale)
        getattr(model, __var.ode_constraints)[k] = scaled_expr == 0


def remove_scaling(model, bounds=None):
    """This method resets the scaling

    :param ConcreteModel model: A pyomo model
    :param tuple bounds: The bounds to be reset

    :return: None

    """
    if not hasattr(model, __var.model_parameter_scaled):
        raise AttributeError('The model is not scaled')
        
    for param in getattr(model, __var.model_parameter).keys():   
        change_value = getattr(model, __var.model_parameter_scaled)[param].value
        getattr(model, __var.model_parameter)[param].set_value(getattr(model, __var.model_parameter_scaled)[param].value)
        
        for k, v in model.odes.items():
            ep_updated_expr = update_expression(v.body, getattr(model, __var.model_parameter_scaled)[param], 1)
            model.odes[k] = ep_updated_expr == 0
        
        del getattr(model, __var.model_parameter_scaled)[param]
        
    if bounds is not None:
        for param, bound in bounds.items():
            if param in getattr(model, __var.model_parameter):
                getattr(model, __var.model_parameter)[param].setlb(bound[0])
                getattr(model, __var.model_parameter)[param].setub(bound[1])
        
    return None
        
def _scale_expression(expr, scale):
    """Method to set up the scaling

    :param expression expr: The expression to scale
    :param dict scale: The mapping of ids to variables

    :return: The scaled expression
    :rtype: expression

    """
    visitor = ScalingVisitor(scale)
    return visitor.dfs_postorder_stack(expr)


def update_expression(expr, replacement_param, change_value):
    """Takes the non-estiambale parameter and replaces it with its intitial
    value

    .. warning::

        This may not be needed and may be deleted

    :param expression expr: The target ode constraint
    :param str replacement_param: The non-estimable parameter to replace
    :param float change_value: The initial value for the above parameter

    :param expression new_expr: updated constraints with the
        desired parameter replaced with a float

    :return: Updated expression
    :rtype: expression

    """
    visitor = ReplacementVisitor()
    visitor.change_replacement(change_value)
    visitor.change_suspect(id(replacement_param))
    new_expr = visitor.dfs_postorder_stack(expr)
    
    return new_expr



"""
Reduced Hessian Generation - 2nd Edition

This module creates the reduced Hessian for use in various KIPET modules
"""
# Standard library imports
import os
from pathlib import Path

# Third party imports
import numpy as np
import pandas as pd
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.environ import Constraint, Param, Set, SolverFactory, Suffix
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix, triu
from scipy.sparse.linalg import spsolve

# KIPET library imports
from kipet.estimability_tools.parameter_handling import set_scaled_parameter_bounds

DEBUG = True

def get_kkt_info(model_object, solver_results):
    """Takes the results object from SolverFactory and uses PyNumero or k_aug to get the jacobian and Hessian
    information as dataframes. This is done in place and does not return anything.

    kkt_data (dict): dictionary with the following structure:

            {
            'J': J,   # Jacobian
            'H': H,   # Hessian
            'var_ind': var_index_names, # Variable index
            'con_ind': con_index_names, # Constraint index
            'duals': duals, # Duals
            }

    :return: None

    """
    kaug = SolverFactory('k_aug')
    
    kaug.options["print_kkt"] = ""
    res = kaug.solve(model_object, tee=True)

    num_vars = res['Problem'][0]['Number of variables']
    num_cons = res['Problem'][0]['Number of constraints']

    kaug_files = Path('GJH')
    
    var_index_names = pd.read_csv(self.sol_files['col'], sep=';', header=None)  # dummy sep
    con_index_names = pd.read_csv(self.sol_files['row'], sep=';', header=None)  # dummy sep

    var_index_names = [var_name for var_name in var_index_names[0]]
    con_index_names = [con_name for con_name in con_index_names[0].iloc[:-1]]
    # con_index_number = {v: k for k, v in enumerate(con_index_names)}

    n = len(var_index_names)
    m = len(con_index_names)

    hess_file = kaug_files.joinpath('H_print.txt')
    hess = pd.read_csv(hess_file, delim_whitespace=True, header=None, skipinitialspace=True)
    
    hess.columns = ['irow', 'jcol', 'vals']
    hess.irow -= 1
    hess.jcol -= 1
    # os.unlink(f'{kaug_files}hess_debug.in')

    jac_file = kaug_files.joinpath('A_print.txt')
    jac = pd.read_csv(jac_file, delim_whitespace=True, header=None, skipinitialspace=True)

    jac.columns = ['irow', 'jcol', 'vals']
    jac.irow -= 1
    jac.jcol -= 1
    # os.unlink(f'{kaug_files}jacobi_debug.in')

    # try:
    #    duals = read_duals(stub + '.sol')
    # except:
    duals = None

    J = coo_matrix((jac.vals, (jac.jcol, jac.irow)), shape=(m, n))
    Hess_coo = coo_matrix((hess.vals, (hess.irow, hess.jcol)), shape=(n, n))
    H = Hess_coo + triu(Hess_coo, 1).T

        

    self.delete_sol_files()

    self.kkt_data = {
        'J': J,
        'H': H,
        'var_ind': var_index_names,
        'con_ind': con_index_names,
        'duals': duals,
    }

    return None


def calculate_reduced_hessian(self, d=None, optimize=False, return_Z=False):
    """Calculate the reduced Hessian
    
    :param dict d: The current parameter values
    :param bool optimize: Option to optimize the problem (if not done already)
    :param bool return_Z: Option to add the Z_mat to the returned values

    .. note::

        If returning the Z_mat, the returned object will be a tuple of the reduced Hessian and the Z matrix.

    :return: The reduced Hessian and optionally the Z matrix
    :rtype: numpy.ndarray
    
    """
    if optimize:
        self.optimize_model(d)

    self.get_kkt_info()
    H = self.kkt_data['H']
    J = self.kkt_data['J']
    var_ind = self.kkt_data['var_ind']
    con_ind_new = self.kkt_data['con_ind']
    duals = self.kkt_data['duals']

    col_ind = [var_ind.index(f'{self.variable_name}[{v}]') for v in self.parameter_set]
    m, n = J.shape

    if self.param_con_method == 'global':

        dummy_constraints = [f'{self.global_constraint_name}[{k}]' for k in self.parameter_set]
        # print(dummy_constraints)
        jac_row_ind = [con_ind_new.index(d) for d in dummy_constraints]
        # duals_imp = [duals[i] for i in jac_row_ind]

        # print(J.shape, len(duals_imp))

        J_c = delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
        row_indexer = SparseRowIndexer(J_c)
        J_f = row_indexer[col_ind]
        J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[])
        J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)
        # print(J_f, J_l)

    elif self.param_con_method == 'fixed':

        jac_row_ind = []
        duals_imp = None

        J_c = J.tocsc()  # delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
        row_indexer = SparseRowIndexer(J_c.T)
        J_f = row_indexer[col_ind].T
        # J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[])
        J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)
        # print(J_f, J_l)

    else:
        None

    r_hess, Z_mat = self._reduced_hessian_matrix(J_f, J_l, H, col_ind)

    if not return_Z:
        return r_hess.todense()
    else:
        return r_hess.todense(), Z_mat


def _reduced_hessian_matrix(F, L, H, col_ind):
    """This calculates the reduced hessian by calculating the null-space based
    on the constraints

    :param csr_matrix F: Rows of the Jacobian related to fixed parameters
    :param csr_matrix L: The remainder of the Jacobian without parameters
    :param csr_matrix H: The sparse Hessian
    :param list col_ind: indices of columns with fixed parameters
    
    :return: sparse version of the reduced Hessian
    :rtype: csr_matrix

    """
    n = H.shape[0]
    n_free = n - F.shape[0]

    X = spsolve(L.tocsc(), -F.tocsc())

    col_ind_left = list(set(range(n)).difference(set(col_ind)))
    col_ind_left.sort()

    Z = np.zeros([n, n_free])
    Z[col_ind, :] = np.eye(n_free)

    if isinstance(X, csc_matrix):
        Z[col_ind_left, :] = X.todense()
    else:
        Z[col_ind_left, :] = X.reshape(-1, 1)

    Z_mat = coo_matrix(np.mat(Z)).tocsr()
    Z_mat_T = coo_matrix(np.mat(Z).T).tocsr()
    Hess = H.tocsr()
    reduced_hessian = Z_mat_T * Hess * Z_mat

    return reduced_hessian, Z_mat


def read_duals(sol_file):
    """Reads the duals from the sol file after solving the problem

    :param str sol_file: The absolute path to the sol file

    :return list duals: The list of duals values taken from the sol file
    
    """
    sol_file_abs = Path(sol_file)

    duals = []
    with sol_file_abs.open() as f:
        lines = f.readlines()

    lines_found = True
    num_of_vars = int(lines[9])

    for ln, line in enumerate(lines):
        line = line.rstrip('\n')
        line = line.lstrip('\t').lstrip(' ')

        if ln >= 12 and ln <= (num_of_vars + 11):
            duals.append(float(line))

    return duals


def delete_from_csr(mat, row_indices=[], col_indices=[]):
    """
    Remove the rows (denoted by ``row_indices``) and columns (denoted by 
    ``col_indices``) from the CSR sparse matrix ``mat``.
    WARNING: Indices of altered axes are reset in the returned matrix

    :param csr_matrix mat: Sparse matrix to delete rows and cols from
    :param list row_indicies: rows to delete
    :param list col_indicies: cols to delete

    :return csr_matrix mat: The sparse matrix with the rows and cols removed

    """
    if not isinstance(mat, csr_matrix):
        raise ValueError("works only for CSR format -- use .tocsr() first")

    rows = []
    cols = []
    if row_indices:
        rows = list(row_indices)
    if col_indices:
        cols = list(col_indices)

    if len(rows) > 0 and len(cols) > 0:
        row_mask = np.ones(mat.shape[0], dtype=bool)
        row_mask[rows] = False
        col_mask = np.ones(mat.shape[1], dtype=bool)
        col_mask[cols] = False
        return mat[row_mask][:, col_mask]
    elif len(rows) > 0:
        mask = np.ones(mat.shape[0], dtype=bool)
        mask[rows] = False
        return mat[mask]
    elif len(cols) > 0:
        mask = np.ones(mat.shape[1], dtype=bool)
        mask[cols] = False
        return mat[:, mask]
    else:
        return mat


class SparseRowIndexer:
    """Class used to splice sparse matrices"""

    def __init__(self, matrix):
        data = []
        indices = []
        indptr = []

        _class = 'csr'
        # print(f'LOOK HERE: {type(matrix)}')
        if isinstance(matrix, csc_matrix):
            _class = 'csc'
        #     print(_class)

        self._class = _class
        # Iterating over the rows this way is significantly more efficient
        # than matrix[row_index,:] and matrix.getrow(row_index)
        for row_start, row_end in zip(matrix.indptr[:-1], matrix.indptr[1:]):
            data.append(matrix.data[row_start:row_end])
            indices.append(matrix.indices[row_start:row_end])
            indptr.append(row_end - row_start)  # nnz of the row

        self.data = np.array(data, dtype=object)
        self.indices = np.array(indices, dtype=object)
        self.indptr = np.array(indptr)
        self.n_columns = matrix.shape[1]

    def __getitem__(self, row_selector):
        data = np.concatenate(self.data[row_selector])
        indices = np.concatenate(self.indices[row_selector])
        indptr = np.append(0, np.cumsum(self.indptr[row_selector]))

        shape = [indptr.shape[0] - 1, self.n_columns]

        if self._class == 'csr':
            return csr_matrix((data, indices, indptr), shape=shape)
        else:
            return csr_matrix((data, indices, indptr), shape=shape).T.tocsc()



"""
Settings for KIPET
"""
# Standard library imports
import ast
from pathlib import Path

# Third party imports
import yaml

# Kipet library imports
from kipet.calculation_tools.helper import AttrDict


class Settings:
    """This is a container for all of the options that can be used in Kipet
    Since it can be confusing due to the large number of options, this should
    make it easier for the user to see everything in one spot.
    
    This class loads the default settings from the settings.yml file located
    in the directory wherever KIPET has been installed. If you really want
    to change this file, go ahead, but make sure that you save a copy of the
    original settings in case something goes wrong.
    
    :Methods:
    
    - :func:`updated_settings`
    
    """

    def __init__(self, category='model'):
        """Settings object initialization that begins each time a
        ReactionModel instance is created.
        
        """
        self._load_settings()
        self._reset_model()

    def __str__(self):

        m = 25

        settings = 'Settings\n\n'

        if hasattr(self, 'general'):
            settings += 'General Settings (general):\n'
            for k, v in self.general.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        if hasattr(self, 'units'):
            settings += '\nUnit Settings (units):\n'
            for k, v in self.units.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        if hasattr(self, 'collocation'):
            settings += '\nCollocation Settings (collocation):\n'
            for k, v in self.collocation.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        if hasattr(self, 'simulator'):
            settings += '\nSimulation Settings (simulator):\n'
            for k, v in self.simulator.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        if hasattr(self, 'variance_estimator'):
            settings += '\nVarianceEstimator Settings (variance_estimator):\n'
            for k, v in self.variance_estimator.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        if hasattr(self, 'parameter_estimator'):
            settings += '\nParameterEstimator Settings (parameter_estimator):\n'
            for k, v in self.parameter_estimator.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        if hasattr(self, 'solver'):
            settings += '\nSolver Settings (solver):\n'
            for k, v in self.solver.items():
                settings += f'{str(k).rjust(m)} : {v}\n'

        return settings

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def _get_file_path():
        """Finds the path to the settings.yml file by referencing the directory
        of the current file
        
        """

        current_dir = Path(__file__).parent
        settings_file = (current_dir / '../general_settings/settings.yml').resolve()

        return settings_file

    def _load_settings(self):
        """Loads the settings and places the data into the correct dict
        structure
        
        """
        settings_file = self._get_file_path()

        with open(settings_file) as file:
            self.cfg = yaml.load(file, Loader=yaml.FullLoader)

            for sub_dict in self.cfg.values():
                for key, value in sub_dict.items():
                    if isinstance(value, bool):
                        continue
                    elif value in ['True', 'False', 'None']:
                        sub_dict[key] = ast.literal_eval(value)
                    elif _is_number(value):
                        if float(value) == int(float(value)):
                            sub_dict[key] = int(float(value))
                        else:
                            sub_dict[key] = float(value)
                    else:
                        None

        return None

    def _reset_model(self):
        """Initializes the settings dicts to their default values
        
        """
        self.collocation = AttrDict(self.cfg['collocation'])
        self.simulator = AttrDict(self.cfg['simulator'])
        self.simulator.update({'solver_opts': AttrDict()})
        self.general = AttrDict(self.cfg['general'])
        self.variance_estimator = AttrDict(self.cfg['variance_estimator'])
        self.variance_estimator.update({'solver_opts': AttrDict()})
        self.parameter_estimator = AttrDict(self.cfg['parameter_estimator'])
        self.parameter_estimator.update({'solver_opts': AttrDict()})
        self.solver = AttrDict(self.cfg['solver'])
        #self.units = AttrDict(self.cfg['units'])

        return None

    def update_settings(self, category, item, value):
        """Sets the default settings to some new value
        
        This allows the user to make permanent changes to the settings file.
        
        .. warning::
            
            Careful! This may result in KIPET not working properly if you make
            a change that is incompatible!
        
        :param str category: The category containing the value to change
        :param str item: The name of the setting to change
        :param str value: The new value for the setting
        
        :return: None
        
        """
        settings_file = self._get_file_path()

        self.cfg[category][item] = value
        with open(settings_file, 'w') as yaml_file:
            yaml_file.write(yaml.dump(self.cfg, default_flow_style=False))

        print(f'Updated default settings:\n\n{category}:\n\t{item}: {value}')
        return None

    @property
    def as_dicts(self):
        
        full_name = {
            'general' : 'General Settings',
            'collocation' : 'Collocation Settings',
            'simulator' : 'Simulation Settings',
            'variance_estimator' : 'Variance Estimator Settings',
            'parameter_estimator' : 'Parameter Estimator Settings',
            'solver' : 'Solver Settings',
        }
        
        keys = ['collocation', 'simulator', 'general', 'variance_estimator',
                'parameter_estimator', 'solver']
        
        nested_dict = {}
        for key in keys:
            nested_dict[key] = (full_name[key], getattr(self, key))
            
        return nested_dict
        


def _is_number(s):
    """ Returns True if the input is a float (number), else False

    :param str s: String that will be checked as to whether it is a number
    
    :return: True or False, depending on whether a number is detected.
    :rtype: bool
    
    """
    try:
        float(s)
        return True
    finally:
        return False



"""
Holds the alternate method used in VarianceEstimator
"""
# Standard library imports
import warnings

# Third party imports
from pyomo.core import log, value
from pyomo.environ import Objective, Param, SolverFactory

# KIPET library imports
from kipet.estimator_tools.results_object import ResultsObject


def run_alternate_method(var_est_object, solver, run_opt_kwargs):
    """Calls the alternative method - Short et al 2020

    This is an improved method for determining the component variances. This method has been removed from the
    VarianceEstimator class for simplification.

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict run_opt_kwargs: The dict of user settings passed on from the ReactionModel

    :return results: The results from the variance estimation
    :rtype: ResultsObject

    """
    # Unpack the keyword arguments
    solver_opts = run_opt_kwargs.pop('solver_opts', dict())
    init_sigmas = run_opt_kwargs.pop('initial_sigmas', float())
    tee = run_opt_kwargs.pop('tee', False)
    tol = run_opt_kwargs.pop('tolerance', 5.0e-5)
    A = run_opt_kwargs.pop('subset_lambdas', None)
    secant_point2 = run_opt_kwargs.pop('secant_point', None)
    individual_species = run_opt_kwargs.pop('individual_species', False)

    # Solver is fixed to ipopt
    solver = 'ipopt'

    nu_squared = var_est_object.solve_max_device_variance(
                                           solver,
                                           tee=tee,
                                           subset_lambdas=A, 
                                           solver_opts=solver_opts)

    second_point = init_sigmas*10
    if secant_point2 is not None:
        second_point = secant_point2

    itersigma = dict()
    itersigma[0] = second_point
    itersigma[1] = init_sigmas
    iterdelta = dict()
    count = 1
    tol = tol
    funcval = 1000
    tee = False
    
    iterdelta[0] = _solve_delta_given_sigma(var_est_object,
                                            solver,
                                            tee=tee, 
                                            subset_lambdas=A, 
                                            solver_opts=solver_opts, 
                                            init_sigmas=itersigma[0])

    while abs(funcval) >= tol:
        print("Overall sigma value at iteration", count, ": ", itersigma[count])
        
        new_delta = _solve_delta_given_sigma(var_est_object,
                                             solver, 
                                             tee=tee, 
                                             subset_lambdas=A, 
                                             solver_opts=solver_opts, 
                                             init_sigmas=itersigma[count])

        print("New delta_sq val: ", new_delta)
        iterdelta[count] = new_delta
        
        def func1(nu_squared, new_delta, init_sigmas):
            sigmult = 0
            nwp = len(var_est_object.model.meas_lambdas)
            for l in var_est_object.model.meas_lambdas:
                for k in var_est_object.comps['unknown_absorbance']:
                   sigmult += value(var_est_object.model.S[l, k])
            funcval = nu_squared - new_delta - init_sigmas*(sigmult/nwp)
            return funcval, sigmult
       
        funcval, sigmult = func1(nu_squared, new_delta, itersigma[count])

        if abs(funcval) <= tol:
            break
        else:
            denom_secant = func1(nu_squared, new_delta, itersigma[count])[0] - func1(nu_squared, iterdelta[count-1], itersigma[count - 1])[0]
            itersigma[count + 1] = itersigma[count] - func1(nu_squared, new_delta, itersigma[count])[0]*((itersigma[count]-itersigma[count-1])/denom_secant)
            
            if itersigma[count + 1] < 0:
                itersigma[count + 1] = -1*itersigma[count + 1]
            count += 1
            
    if individual_species:
        print("Solving for individual species' variance based on the obtained delta")
        max_likelihood_val, sigma_vals, stop_it, results = \
            _solve_sigma_given_delta(var_est_object,
                                          solver, 
                                          subset_lambdas= A, 
                                          solver_opts=solver_opts, 
                                          tee=tee, 
                                          delta=new_delta)
    else:
        print("The overall model variance is: ", itersigma[count])
        
        sigma_vals = {}
        for k in var_est_object.comps['unknown_absorbance']:
            sigma_vals[k] = abs(itersigma[count])
        
    print(f'sigma_vals: {sigma_vals}')
    results = ResultsObject()
    results.load_from_pyomo_model(var_est_object.model)
    results.sigma_sq = sigma_vals
    results.sigma_sq['device'] = new_delta

    return results


def run_direct_sigmas_method(var_est_object, solver, run_opt_kwargs, fixed=False):
    """"Calls the direct sigmas method

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used
    :param dict run_opt_kwargs: The dict of user settings passed on from the ReactionModel

    :return results: The results from the variance estimation
    :rtype: ResultsObject

    """
    solver_opts = run_opt_kwargs.pop('solver_opts', dict())
    tee = run_opt_kwargs.pop('tee', False)
    A = run_opt_kwargs.pop('freq_subset_lambdas', None)
    fixed_device_var = run_opt_kwargs.pop('fixed_device_variance', None)
    device_range = run_opt_kwargs.pop('device_range', None)
    num_points = run_opt_kwargs.pop('num_points', None)
    
    print("Solving for sigmas assuming known device variances")
    
    print('Device range')
    print(device_range)
    
    if device_range:
        if not isinstance(device_range, tuple):
            print("device_range is of type {}".format(type(device_range)))
            print("It should be a tuple")
            raise Exception
        elif device_range[0] > device_range[1]:
            print("device_range needs to be arranged in order from lowest to highest")
            raise Exception
        else:
            print("Device range means that we will solve iteratively for different delta values in that range")
    
    else:
        fixed = True
        if fixed_device_var is None:
            raise ValueError("If using fixed variance, this needs to be provided.")
    
    if device_range and not num_points:
        print("Need to specify the number of points that we wish to evaluate in the device range")
    if not num_points:
        pass
    elif not isinstance(num_points, int):  
        print("num_points needs to be an integer!")
        raise Exception
        
    if not device_range and not num_points:
        
        print("assessing for the value of delta provided")
        if not fixed_device_var:
            print("If iterative method not selected then need to provide fixed device variance (delta**2)")
            raise Exception
        else:
            if not isinstance(fixed_device_var, float):
                raise Exception("fixed device variance needs to be of type float")
    
    if not fixed:
        
        results_sigmas_dict = {}
        
        dist = abs((device_range[1] - device_range[0])/num_points)
        
        max_likelihood_vals = []
        delta_vals = []
        iteration_counter = []
        delta = device_range[0]
        
        count = 0
        
        print('*** Starting Variance Iterations ***')
       
        while delta < device_range[1]:
            
            results_sigmas_dict[count] = {}
            
            print(f"Iteration: {count}\tdelta_sq: {delta}")
            max_likelihood_val, sigma_vals, stop_it, results = \
                _solve_sigma_given_delta(var_est_object,
                                         solver, 
                                         subset_lambdas=A, 
                                         solver_opts=solver_opts, 
                                         tee=tee,
                                         delta=delta
                                    )
            
            if max_likelihood_val >= 5000:
                max_likelihood_vals.append(5000)
                delta_vals.append(log(delta))
                iteration_counter.append(count)
                max_likelihood_vals.append(max_likelihood_val)
                delta_vals.append(log(delta))
            else:
                iteration_counter.append(count)
                
            results_sigmas_dict[count]['delta'] = delta
            results_sigmas_dict[count]['simgas'] = sigma_vals
            #results_sigmas_dict[count]['results'] = results
                
            delta = delta + dist
            count += 1 
            
        return results_sigmas_dict

    else:
        # The optimization will be conducted at the fixed value for delta
        
        max_likelihood_val, sigma_vals, stop_it, results = \
            _solve_sigma_given_delta(var_est_object,
                                     solver, 
                                     subset_lambdas=A, 
                                     solver_opts=solver_opts, 
                                     tee=tee,
                                     delta=fixed_device_var
                                )
            
        delta = fixed_device_var
   
        results = ResultsObject()
        results.load_from_pyomo_model(var_est_object.model)
        results.sigma_sq = sigma_vals
        results.sigma_sq['device'] = delta
            
        return results


def _solve_delta_given_sigma(var_est_object, solver, **kwds):
    """Solves the delta with provided variances

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict kwds: The dict of user settings passed on from the ReactionModel

    :return results: The results from the variance estimation
    :rtype: ResultsObject

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    set_A = kwds.pop('subset_lambdas', list())
    profile_time = kwds.pop('profile_time', False)
    sigmas = kwds.pop('init_sigmas', dict() or float() )

    sigmas_sq = dict()
    if not set_A:
        set_A = var_est_object.model.meas_lambdas
    
    if isinstance(sigmas, float):
        for k in var_est_object.comps['unknown_absorbance']:
            sigmas_sq[k] = sigmas
    
    elif isinstance(sigmas, dict):
        keys = sigmas.keys()
        for k in var_est_object.comps['unknown_absorbance']:
            if k not in keys:
                sigmas_sq[k] = sigmas
       
    print("Solving delta from given sigmas\n")
    print(sigmas_sq)
   
    var_est_object._warn_if_D_negative()  
   
    obj = 0.0
    ntp = len(var_est_object.model.times_spectral)
    nwp = len(var_est_object.model.meas_lambdas) 
    inlog = 0
    nc = len(var_est_object.comps['unknown_absorbance'])
    for t in var_est_object.model.times_spectral:
        for l in set_A:
            D_bar = sum(var_est_object.model.C[t, k] * var_est_object.model.S[l, k] for k in var_est_object.comps['unknown_absorbance'])
            #D_bar = sum(var_est_object.model.C[t, k] * var_est_object.model.S[l, k] for k in var_est_object._sublist_components)
            inlog += (var_est_object.model.D[t, l] - D_bar)**2
    # Orig had sublist in both parts - is this an error?

    # Concentration - correct
    # Move this to objectives module
    for t in var_est_object.model.times_spectral:
        for k in var_est_object.comps['unknown_absorbance']:
            #obj += conc_objective(var_est_object.model, sigma=sigmas_sq)
            obj += 0.5*((var_est_object.model.C[t, k] - var_est_object.model.Z[t, k])**2)/(sigmas_sq[k])
            
    obj += (nwp)*log((inlog)+1e-12)     
    var_est_object.model.init_objective = Objective(expr=obj)
    
    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
    solver_results = opt.solve(var_est_object.model,
                               tee=tee,
                               report_timing=profile_time)
    residuals = (value(var_est_object.model.init_objective))
    
    print("residuals: ", residuals)

    for k, v in var_est_object.model.P.items():
        print(k, v.value)
        
    etaTeta = 0
    for t in var_est_object.model.times_spectral:
        for l in set_A:
            D_bar = sum(value(var_est_object.model.C[t, k]) * value(var_est_object.model.S[l, k]) for k in var_est_object.comps['unknown_absorbance'])
            etaTeta += (value(var_est_object.model.D[t, l]) - D_bar)**2
    
    deltasq = etaTeta/(ntp*nwp)  
    var_est_object.model.del_component('init_objective')
   
    return deltasq

def _solve_sigma_given_delta(var_est_object, solver, **kwds):
    """Solves the delta (device variance) with provided variances

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict kwds: The dict of user settings passed on from the ReactionModel

    :return residuals: The results from the variance estimation
    :return variances_dict: dictionary containing the model variance values
    :return stop_it: boolean indicator showing whether no solution was found (True) or if there is a solution (False)
    :return solver_results: dictionary containing solver options for IPOPT

    :rtype: ResultsObject

    :Keyword Args:

        - delta (float): the device variance
        - tee (bool,optional): flag to tell the optimizer whether to stream output to the terminal or not
        - profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model. Default False
        - subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem (Weifeng paper). Default all wavelengths.
        - solver_opts (dict, optional): dictionary containing solver options for IPOPT

    """   
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    set_A = kwds.pop('subset_lambdas', list())
    profile_time = kwds.pop('profile_time', False)
    delta_sq = kwds.pop('delta', dict())
    #species_list = kwds.pop('subset_components', None)

    model = var_est_object.model.clone()

    if not set_A:
        set_A = var_est_object.model.meas_lambdas
        
    # if not hasattr(var_est_object, '_sublist_components'):
    #     list_components = []
    #     if species_list is None:
    #         list_components = [k for k in var_est_object._mixture_components]
            
    #     else:
    #         for k in species_list:
    #             if k in var_est_object._mixture_components:
    #                 list_components.append(k)
    #             else:
    #                 warnings.warn("Ignored {} since is not a mixture component of the model".format(k))

    #     var_est_object._sublist_components = list_components
    
    var_est_object._warn_if_D_negative()  
    ntp = len(var_est_object.model.times_spectral)
    obj = 0.0
   
    for t in var_est_object.model.times_spectral:
        for l in set_A:
            D_bar = sum(var_est_object.model.C[t, k] * var_est_object.model.S[l, k] for k in var_est_object.comps['unknown_absorbance'])
            obj += 0.5/delta_sq*(var_est_object.model.D[t, l] - D_bar)**2

    inlog = {k: 0 for k in var_est_object.comps['unknown_absorbance']}
    var_est_object.model.eps = Param(initialize = 1e-8)  
    
    variances_dict = {k: 0 for k in var_est_object.comps['unknown_absorbance']}
            
    for t in var_est_object.model.times_spectral:
        for k in var_est_object.comps['unknown_absorbance']:
            inlog[k] += ((var_est_object.model.C[t, k] - var_est_object.model.Z[t, k])**2)
    
    for k in var_est_object.comps['unknown_absorbance']:
        obj += 0.5*ntp*log(inlog[k]/ntp + var_est_object.model.eps)
    
    var_est_object.model.init_objective = Objective(expr=obj)

    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
    try:
        solver_results = opt.solve(var_est_object.model,
                               tee=tee,
                               report_timing=profile_time)

        residuals = (value(var_est_object.model.init_objective))
        for t in var_est_object.model.times_spectral:
            for k in var_est_object.comps['unknown_absorbance']:
                variances_dict[k] += 1 / ntp*((value(var_est_object.model.C[t, k]) - value(var_est_object.model.Z[t, k]))**2)
        
        print("Variances")
        for k in var_est_object.comps['unknown_absorbance']:
            print(k, variances_dict[k])
        
        print("Parameter estimates")
        for k, v in var_est_object.model.P.items():
            print(k, v.value)
           
        stop_it = False
        
    except:
        print("FAILED AT THIS ITERATION")
        variances_dict = None
        residuals = 0
        stop_it = True
        solver_results = None
        for k, v in var_est_object.model.P.items():
            print(k, v.value)
        var_est_object.model = model
        for k, v in var_est_object.model.P.items():
            print(k, v.value)
    var_est_object.model.del_component('init_objective')
    var_est_object.model.del_component('eps')

    return residuals, variances_dict, stop_it, solver_results
